import math
import warnings
import numpy as np
import pandas.api.types as pat
from sklearn.linear_model import LinearRegression as _LinearRegression, Lasso as _Lasso, Ridge as _Ridge

from inferelator.regression import base_regression
from inferelator.distributed.inferelator_mp import MPControl
from inferelator import utils

_DEFAULT_ALPHAS = np.insert(np.logspace(-2, 1, 20), 0, 0)
_DEFAULT_NUM_SUBSAMPLES = 20
_DEFAULT_THRESHOLD = 0.05
_DEFAULT_SEED = 42
_DEFAULT_METHOD = 'lasso'
_DEFAULT_PARAMS = {"max_iter": 2000}


def lasso(x, y, alpha, **kwargs):
    return _regress(x, y, alpha, _Lasso, **kwargs)


def ridge(x, y, alpha, ridge_threshold=1e-2, **kwargs):
    betas = _regress(x, y, alpha, _Ridge, **kwargs).flatten()
    betas[betas < ridge_threshold] = 0.
    return betas


def _regress(x, y, alpha, regression, **kwargs):
    if alpha == 0:
        return _LinearRegression(normalize=False).fit(x, y).coef_
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return regression(alpha=alpha, fit_intercept=False, normalize=False, **kwargs).fit(x, y).coef_


def stars_model_select(x, y, alphas, threshold=_DEFAULT_THRESHOLD, num_subsamples=_DEFAULT_NUM_SUBSAMPLES,
                       random_seed=_DEFAULT_SEED, method='lasso', **kwargs):
    """
    Model using StARS (Stability Approach to Regularization Selection) for model selection

    :param x:
    :param y:
    :param alphas:
    :param threshold:
    :param num_subsamples:
    :param random_seed:
    :param method:
    :param kwargs:
    :return:
    """

    if method.lower() == 'lasso':
        _regress_func = lasso
    elif method.lower() == 'ridge':
        _regress_func = ridge
    else:
        raise ValueError("Method must be 'lasso' or 'ridge'")

    # Number of obs
    n, k = x.shape

    if n < num_subsamples:
        msg = "Subsamples ({ns}) for StARS is larger than the number of samples ({n})".format(ns=num_subsamples, n=n)
        raise ValueError(msg)

    # Calculate the number of obs per subsample
    b = math.floor(n / num_subsamples)

    # Make an index for subsampling
    idx = _make_subsample_idx(n, b, num_subsamples, random_seed=random_seed)

    # Calculate betas for stability selection
    betas = {a: [] for a in alphas}
    for sample in range(num_subsamples):
        # Sample and put into column-major (the coordinate descent implementation in sklearn wants that order)
        x_samp = np.asarray(x[idx == sample, :], order='F')
        y_samp = y[idx == sample]

        for a in alphas:
            betas[a].append(_regress_func(x_samp, y_samp, a, **kwargs))

    # Calculate edge stability
    stabilities = {a: _calculate_stability(betas[a]) for a in alphas}

    # Calculate monotonic increasing (as alpha decreases) mean edge stability
    alphas = np.sort(alphas)[::-1]
    total_instability = np.array([np.mean(stabilities[a]) for a in alphas])

    for i in range(1, len(total_instability)):
        if total_instability[i] < total_instability[i - 1]:
            total_instability[i] = total_instability[i - 1]

    threshold_alphas = np.array(alphas)[total_instability < threshold]
    selected_alpha = np.min(threshold_alphas) if len(threshold_alphas) > 0 else alphas[0]

    refit_betas = _regress_func(x, y, selected_alpha, **kwargs)
    beta_nonzero = _make_bool_matrix(refit_betas)

    if beta_nonzero.sum() == 0:
        return dict(pp=np.repeat(True, k).tolist(),
                    betas=np.zeros(k),
                    betas_resc=np.zeros(k))
    else:
        x = x[:, beta_nonzero]
        utils.make_array_2d(y)
        betas = base_regression.recalculate_betas_from_selected(x, y)
        betas_resc = base_regression.predict_error_reduction(x, y, betas)

        return dict(pp=beta_nonzero,
                    betas=betas,
                    betas_resc=betas_resc)


def _calculate_stability(edges):
    """

    :param edges:
    :return:
    """

    edge_sum = np.zeros(edges[0].shape, dtype=int)
    for e in edges:
        edge_sum += _make_bool_matrix(e)

    edge_sum = edge_sum / len(edges)
    return 2 * edge_sum * (1 - edge_sum)


def _make_bool_matrix(edge_matrix):
    if pat.is_float_dtype(edge_matrix.dtype):
        return np.abs(edge_matrix) > np.finfo(dtype=edge_matrix.dtype).eps
    else:
        return edge_matrix != 0


def _make_subsample_idx(n, b, num_subsamples, random_seed=42):
    # Partition all samples into num_subsamples groups
    subsample_index = np.zeros((n,), dtype=np.int8)
    for i in range(b):
        start, stop = i * num_subsamples, min((i + 1) * num_subsamples, n)
        subsample_index[start:stop] = list(range(num_subsamples))[0:stop - start]

    np.random.RandomState(random_seed).shuffle(subsample_index)

    return subsample_index


class StARS(base_regression.BaseRegression):

    def __init__(self, X, Y, random_seed, alphas=_DEFAULT_ALPHAS, num_subsamples=_DEFAULT_NUM_SUBSAMPLES,
                 method=_DEFAULT_METHOD, parameters=None):
        self.random_seed = random_seed
        self.alphas = alphas
        self.num_subsamples = num_subsamples
        self.method = method

        self.params = parameters if parameters is not None else {}

        super(StARS, self).__init__(X, Y)

    def regress(self):
        """
        Execute StARS

        :return: list
            Returns a list of regression results that base_regression's pileup_data can process
        """

        if MPControl.is_dask():
            from inferelator.distributed.dask_functions import lasso_stars_regress_dask
            return lasso_stars_regress_dask(self.X, self.Y, self.alphas, self.num_subsamples, self.random_seed,
                                            self.method, self.params, self.G, self.genes)

        def regression_maker(j):
            level = 0 if j % 100 == 0 else 2
            utils.Debug.allprint(base_regression.PROGRESS_STR.format(gn=self.genes[j], i=j, total=self.G), level=level)

            data = stars_model_select(self.X.values,
                                      utils.scale_vector(self.Y.get_gene_data(j, force_dense=True).flatten()),
                                      self.alphas,
                                      method=self.method,
                                      num_subsamples=self.num_subsamples,
                                      random_seed=self.random_seed,
                                      **self.params)
            data['ind'] = j
            return data

        return MPControl.map(regression_maker, range(self.G), tell_children=False)


class StARSWorkflowMixin(base_regression._RegressionWorkflowMixin):
    """
    Stability Approach to Regularization Selection (StARS)-LASSO. StARS-Ridge is implemented on an experimental basis.

    https://arxiv.org/abs/1006.3316
    https://doi.org/10.1016/j.immuni.2019.06.001
    """

    sklearn_params = _DEFAULT_PARAMS
    alphas = _DEFAULT_ALPHAS
    regress_method = _DEFAULT_METHOD
    num_subsamples = _DEFAULT_NUM_SUBSAMPLES

    def __init__(self, *args, **kwargs):
        self.sklearn_params = {}
        super(StARSWorkflowMixin, self).__init__(*args, **kwargs)

    def set_regression_parameters(self, alphas=None, num_subsamples=None, method=None, **kwargs):
        """
        Set regression parameters for StARS-LASSO

        :param alphas: A list of alpha (L1 term) values to search. Defaults to logspace between 0. and 10.
        :type alphas: list(float)
        :param num_subsamples: The number of groups to break data into. Defaults to 20.
        :type num_subsamples: int
        :param method: The model to use. Can choose from 'lasso' or 'ridge'. Defaults to 'lasso'.
            If 'ridge' is set, ridge_threshold should also be passed. Any value below ridge_threshold will be set to 0.
        :type method: str
        :param kwargs: Any additional arguments will be passed to the LASSO or Ridge scikit-learn object at
            instantiation
        :type kwargs: any
        """

        self.sklearn_params.update(kwargs)
        self.alphas = alphas if alphas is not None else self.alphas
        self.num_subsamples = num_subsamples if num_subsamples is not None else self.num_subsamples
        self.regress_method = method if method is not None else self.regress_method

    def run_regression(self):
        MPControl.sync_processes("pre_stars")

        betas, resc_betas = StARS(self.design, self.response, self.random_seed, alphas=self.alphas,
                                  method=self.regress_method,
                                  num_subsamples=self.num_subsamples,
                                  parameters=self.sklearn_params).run()

        MPControl.sync_processes("post_stars")

        return [betas], [resc_betas]


class StARSWorkflowByTaskMixin(base_regression._MultitaskRegressionWorkflowMixin, StARSWorkflowMixin):
    """
    Stability Approach to Regularization Selection (StARS)-LASSO. StARS-Ridge is implemented on an experimental basis.

    https://arxiv.org/abs/1006.3316
    https://doi.org/10.1016/j.immuni.2019.06.001
    """

    def run_bootstrap(self, bootstrap_idx):
        betas, betas_resc = [], []

        # Select the appropriate bootstrap from each task and stash the data into X and Y
        for k in range(self._n_tasks):
            x = self._task_design[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])
            y = self._task_response[k].get_bootstrap(self._task_bootstraps[k][bootstrap_idx])

            MPControl.sync_processes(pref="sk_pre")

            utils.Debug.vprint('Calculating task {k} betas using StARS'.format(k=k), level=0)
            t_beta, t_br = StARS(x, y, self.random_seed, alphas=self.alphas,
                                 method=self.regress_method,
                                 num_subsamples=self.num_subsamples,
                                 parameters=self.sklearn_params).run()
            betas.append(t_beta)
            betas_resc.append(t_br)

        return betas, betas_resc



import pandas as pd
import numpy as np

from inferelator import utils
from inferelator.utils import Validator as check

DEFAULT_CV_AXIS = 0
DEFAULT_SEED = 2001


class ManagePriors(object):
    """
    The ManagePriors class has the functions to manipulate prior and gold standard data which are called from workflow
    This filters, aligns, crossvalidates, shuffles, etc.
    """

    @staticmethod
    def validate_priors_gold_standard(priors_data, gold_standard):
        """
        Validate the priors and the gold standard, then pass them through

        :param priors_data: pd.DataFrame [G x K]
            Prior data
        :param gold_standard: pd.DataFrame [G x K]
            Gold standard data
        :return priors, gold_standard: pd.DataFrame [G x K], pd.DataFrame [G x K]
        """

        try:
            check.index_values_unique(priors_data.index)
        except ValueError as v_err:
            utils.Debug.vprint("Duplicate gene(s) in prior index", level=0)
            utils.Debug.vprint("\t" + str(v_err), level=0)

        try:
            check.index_values_unique(priors_data.columns)
        except ValueError as v_err:
            utils.Debug.vprint("Duplicate tf(s) in prior index", level=0)
            utils.Debug.vprint("\t" + str(v_err), level=0)

        try:
            check.index_values_unique(gold_standard.index)
        except ValueError as v_err:
            utils.Debug.vprint("Duplicate gene(s) in gold standard index", level=0)
            utils.Debug.vprint("\t" + str(v_err), level=0)

        try:
            check.index_values_unique(gold_standard.columns)
        except ValueError as v_err:
            utils.Debug.vprint("Duplicate tf(s) in gold standard index", level=0)
            utils.Debug.vprint("\t" + str(v_err), level=0)

        return priors_data, gold_standard

    @staticmethod
    def cross_validate_gold_standard(priors_data, gold_standard, cv_split_axis, cv_split_ratio, random_seed):
        """
        Sample the gold standard for crossvalidation, and then remove the new gold standard from the priors (if split
        on an axis)

        :param priors_data: pd.DataFrame [G x K]
            Prior data
        :param gold_standard: pd.DataFrame [G x K]
            Gold standard data
        :param cv_split_ratio: float
            The proportion of the priors that should go into the gold standard
        :param cv_split_axis: int
            Splits on rows (when 0), columns (when 1), or on flattened individual data points (when None)
            Note that if this is None, the returned gold standard will be the same as all_data, and the priors will have
            half of the data points of all_data
        :param random_seed: int
            Random seed
        :return priors_data, gold_standard: pd.DataFrame [G x K], pd.DataFrame [G x K]
        """

        assert check.argument_enum(cv_split_axis, (0, 1), allow_none=True)
        assert check.argument_numeric(cv_split_ratio, low=0, high=1)

        if cv_split_axis == 1:
            utils.Debug.vprint("Selecting cv_split_axis of 1 is possible but a very bad idea", level=1)

        utils.Debug.vprint("Resampling GS ({gs}) for crossvalidation".format(gs=gold_standard.shape), level=0)
        gs_to_prior, gold_standard = ManagePriors._split_for_cv(gold_standard, cv_split_ratio, split_axis=cv_split_axis,
                                                                seed=random_seed)

        # If the priors are split on an axis, remove circularity
        if cv_split_axis is not None:
            priors_data, gold_standard = ManagePriors._remove_prior_circularity(priors_data, gold_standard,
                                                                                split_axis=cv_split_axis)
        else:
            if priors_data is not None:
                utils.Debug.vprint("Existing prior is being replaced with a downsampled gold standard")
            priors_data = gs_to_prior

        utils.Debug.vprint("CV prior {pr} and gold standard {gs}".format(pr=priors_data.shape,
                                                                         gs=gold_standard.shape), level=0)

        return priors_data, gold_standard

    @staticmethod
    def filter_priors_to_genes(priors_data, gene_list):

        if len(gene_list) == 0:
            raise ValueError("Filtering to a list of 0 genes is not valid")

        if len(priors_data.index) == 0:
            raise ValueError("Filtering a prior matrix of 0 genes is not valid")

        try:
            priors_data = ManagePriors._filter_df_index(priors_data, gene_list)
        except ValueError as err:
            err = str(err) + " when filtering priors for gene list. "
            err += " Prior matrix genes: " + str(priors_data.index[0]) + "..."
            err += " Gene list genes: " + gene_list[0]
            raise ValueError(err)

        return priors_data

    @staticmethod
    def _filter_df_index(data_frame, index_list):
        new_index = data_frame.index.intersection(index_list)

        if len(new_index) == 0:
            raise ValueError("Filtering results in 0-length index")

        return data_frame.loc[new_index, :]

    @staticmethod
    def filter_to_tf_names_list(priors_data, tf_names):
        """
        Filter the priors the intersection with a provided list of regulators
        :param priors_data: pd.DataFrame [G x K]
            Prior data
        :param tf_names: list [k]
            List of regulators to restrict the modeling to
        :return priors_data: pd.DataFrame [G x k]
            Filtered priors on regulators
        """

        tf_keepers = pd.Index(tf_names).intersection(pd.Index(priors_data.columns))
        priors_data = priors_data.loc[:, tf_keepers]

        utils.Debug.vprint("Filtered to {tfn} TFs from the TF name list".format(tfn=len(tf_keepers)), level=1)

        if priors_data.shape[1] == 0:
            raise ValueError("Prior regulators and regulator list regulators have no overlap")

        return priors_data

    @staticmethod
    def align_priors_to_expression(priors_data, gene_list):
        """
        Make sure that the priors align to the expression matrix and fill priors that are created with 0s
        :param priors_data: pd.DataFrame [G x K]
            Prior data
        :param gene_list: pd.Index [G]
            Expression matrix genes
        :return priors_data:
            Returns priors_data where genes match expression matrix genes
        """

        if len(priors_data.index.intersection(gene_list)) == 0:
            err = "Prior genes and expression matrix genes have no overlap."
            if len(priors_data.index) == 0:
                err += " (Prior matrix has no genes"
            else:
                e_genes = map(str, priors_data.index[0:min(len(priors_data.index), 5)])
                err += " (Prior genes: " + " ".join(e_genes) + "..."
            if len(gene_list) == 0:
                err += " Expression matrix has no genes)"
            else:
                e_genes = map(str, gene_list[0:min(len(gene_list), 5)])
                err += " Expression matrix genes: " + " ".join(e_genes) + ")"

            raise ValueError(err)

        return priors_data.reindex(index=gene_list).fillna(value=0)

    @staticmethod
    def shuffle_priors(priors_data, shuffle_prior_axis, random_seed):
        """
        Shuffle the labels on the priors on a specific axis
        :param priors_data: pd.DataFrame [G x K]
            Prior data
        :param shuffle_prior_axis: int
            Axis to shuffle. 0 is genes, 1 is regulators, None is skip shuffling.
        :param random_seed: int
            Random seed
        :return priors_data:
            Returns priors_data the data has been shuffled on a specific axis
        """

        assert check.argument_enum(shuffle_prior_axis, [0, 1], allow_none=True)

        if shuffle_prior_axis is None:
            return priors_data
        elif shuffle_prior_axis == 0:
            # Shuffle index (genes) in the priors_data
            utils.Debug.vprint("Randomly shuffling prior [{sh}] gene data".format(sh=priors_data.shape))
            prior_index = priors_data.index.tolist()
            priors_data = priors_data.sample(frac=1, axis=0, random_state=random_seed)
            priors_data.index = prior_index
        elif shuffle_prior_axis == 1:
            # Shuffle columns (TFs) in the priors_data
            utils.Debug.vprint("Randomly shuffling prior [{sh}] TF data".format(sh=priors_data.shape))
            prior_index = priors_data.columns.tolist()
            priors_data = priors_data.sample(frac=1, axis=1, random_state=random_seed)
            priors_data.columns = prior_index

        return priors_data

    @staticmethod
    def _split_for_cv(all_data, split_ratio, split_axis=DEFAULT_CV_AXIS, seed=DEFAULT_SEED):
        """
        Take a dataframe and split it according to split_ratio on split_axis into two new dataframes. This is for
        crossvalidation splits of a gold standard.

        :param all_data: pd.DataFrame [G x K]
            Existing prior or gold standard data
        :param split_ratio: float
            The proportion of the priors that should go into the gold standard
        :param split_axis: int
            Splits on rows (when 0), columns (when 1), or on flattened individual data points (when None)
            Note that if this is None, the returned gold standard will be the same as all_data, and the priors will have
            half of the data points of all_data
        :param seed: int
            Seed for the random generator
        :return prior_data, gold_standard: pd.DataFrame [G/2 x K], pd.DataFrame [G/2 x K]
            Returns a new prior and gold standard by splitting the old one in half
        """

        assert check.argument_numeric(split_ratio, 0, 1)
        assert check.argument_enum(split_axis, [0, 1], allow_none=True)

        # Split the priors into gold standard based on axis (flatten if axis=None)
        if split_axis is None:
            priors_data, _ = ManagePriors._split_flattened(all_data, split_ratio, seed=seed)
            gold_standard = all_data
        else:
            priors_data, gold_standard = ManagePriors._split_axis(all_data, split_ratio, axis=split_axis, seed=seed)

        return priors_data, gold_standard

    @staticmethod
    def _remove_prior_circularity(priors, gold_standard, split_axis=DEFAULT_CV_AXIS):
        """
        Remove all axis labels that occur in the gold standard from the prior
        :param priors: pd.DataFrame [M x N]
        :param gold_standard: pd.DataFrame [m x n]
        :param split_axis: int (0,1)
        :return new_priors: pd.DataFrame [M-m x N]
        :return gold_standard: pd.DataFrame [m x n]
        """

        assert check.argument_enum(split_axis, [0, 1])

        new_priors = priors.drop(gold_standard.axes[split_axis], axis=split_axis, errors='ignore')

        return new_priors, gold_standard

    @staticmethod
    def _split_flattened(data, split_ratio, seed=DEFAULT_SEED):
        """
        Instead of splitting by axis labels, split edges and ignore axes
        :param data: pd.DataFrame [M x N]
        :param split_ratio: float
        :param seed:
        :return priors_data: pd.DataFrame [M x N]
        :return gold_standard: pd.DataFrame [M x N]
        """

        assert check.argument_numeric(split_ratio, 0, 1)

        pc = np.sum(data.values != 0)
        gs_count = int(split_ratio * pc)
        idx = ManagePriors._make_shuffled_index(pc, seed=seed)

        pr_idx = data.values[data.values != 0].copy()
        gs_idx = data.values[data.values != 0].copy()

        pr_idx[idx[0:gs_count]] = 0
        gs_idx[idx[gs_count:]] = 0

        gs = data.values.copy()
        pr = data.values.copy()

        gs[gs != 0] = gs_idx
        pr[pr != 0] = pr_idx

        priors_data = pd.DataFrame(pr, index=data.index, columns=data.columns)
        gold_standard = pd.DataFrame(gs, index=data.index, columns=data.columns)

        return priors_data, gold_standard

    @staticmethod
    def _split_axis(priors, split_ratio, axis=DEFAULT_CV_AXIS, seed=DEFAULT_SEED):
        """
        Split by axis labels on the chosen axis
        :param priors: pd.DataFrame [M x N]
        :param split_ratio: float
        :param axis: [0, 1]
        :param seed:
        :return:
        """

        assert check.argument_numeric(split_ratio, 0, 1)
        assert check.argument_enum(axis, [0, 1])

        pc = priors.shape[axis]
        gs_count = int((1 - split_ratio) * pc)
        idx = ManagePriors._make_shuffled_index(pc, seed=seed)

        if axis == 0:
            axis_idx = priors.index
        elif axis == 1:
            axis_idx = priors.columns
        else:
            raise ValueError("Axis can only be 0 or 1")

        pr_idx = axis_idx[idx[0:gs_count]]
        gs_idx = axis_idx[idx[gs_count:]]

        priors_data = priors.drop(gs_idx, axis=axis)
        gold_standard = priors.drop(pr_idx, axis=axis)

        return priors_data, gold_standard

    @staticmethod
    def _make_shuffled_index(idx_len, seed=DEFAULT_SEED):
        idx = list(range(idx_len))
        np.random.RandomState(seed=seed).shuffle(idx)
        return idx



import pandas as pd
import pandas.api.types as pat
import numpy as np
import os
import copy as cp
import anndata

from inferelator.utils.data import InferelatorData
from inferelator.utils.debug import Debug
from inferelator.preprocessing.metadata_parser import MetadataHandler

DEFAULT_PANDAS_TSV_SETTINGS = dict(sep="\t", index_col=0, header=0)
DEFAULT_METADATA = "branching"

_TENX_MTX = ("matrix.mtx.gz", "matrix.mtx")
_TENX_BARCODES = ("barcodes.tsv.gz", "barcodes.tsv")
_TENX_FEATURES = ("features.tsv.gz", "genes.tsv")


class InferelatorDataLoader(object):
    input_dir = None
    _file_format_settings = None

    def __init__(self, input_dir, file_format_settings=None):
        self.input_dir = input_dir
        self._file_format_settings = file_format_settings

    def load_data_h5ad(self, h5ad_file, meta_data_file=None, meta_data_handler=DEFAULT_METADATA, gene_data_file=None,
                       gene_name_column=None, use_layer=None):

        data = anndata.read_h5ad(self.input_path(h5ad_file))

        if meta_data_file is None and data.obs.shape[1] > 0:
            meta_data = None
        else:
            meta_data = self.load_metadata_tsv(meta_data_file, data.obs_names, meta_data_handler=meta_data_handler)

        gene_metadata = self.load_gene_metadata_tsv(gene_data_file, gene_name_column)

        if use_layer is not None and use_layer not in data.layers:
            msg = "Layer {lay} is not in {f}".format(lay=use_layer, f=h5ad_file)
            raise ValueError(msg)

        # Build an InferelatorData object from a layer
        elif use_layer is not None:
            data = InferelatorData(data.layers[use_layer],
                                   gene_names=data.var_names,
                                   sample_names=data.obs_names,
                                   meta_data=pd.concat((data.obs, meta_data), axis=1),
                                   gene_data=pd.concat((data.var, gene_metadata), axis=1))

        # Build an InferelatorData object from everything
        else:
            data = InferelatorData(data,
                                   meta_data=meta_data,
                                   gene_data=gene_metadata)

        # Make sure bytestrings are decoded
        _safe_dataframe_decoder(data.gene_data)
        _safe_dataframe_decoder(data.meta_data)

        self._check_loaded_data(data, filename=h5ad_file)
        return data

    def load_data_mtx(self, mtx_file, mtx_obs=None, mtx_feature=None, meta_data_file=None,
                      meta_data_handler=DEFAULT_METADATA, gene_data_file=None, gene_name_column=None):

        data = anndata.read_mtx(self.input_path(mtx_file))
        row_names = self._load_list_from_file(self.input_path(mtx_obs)) if mtx_obs is not None else None
        col_names = self._load_list_from_file(self.input_path(mtx_feature)) if mtx_feature is not None else None

        meta_data = self.load_metadata_tsv(meta_data_file, data.obs_names, meta_data_handler=meta_data_handler)
        gene_metadata = self.load_gene_metadata_tsv(gene_data_file, gene_name_column)

        data = InferelatorData(data,
                               meta_data=meta_data,
                               gene_data=gene_metadata,
                               sample_names=row_names,
                               gene_names=col_names)

        return data

    def load_data_hdf5(self, hdf5_file, use_layer=None, meta_data_file=None, meta_data_handler=DEFAULT_METADATA,
                       gene_data_file=None, gene_name_column=None, transpose_expression_data=False):

        data = pd.HDFStore(self.input_path(hdf5_file), mode='r')
        data = data[data.keys()[0]] if use_layer is None else data[use_layer]

        meta_data = self.load_metadata_tsv(meta_data_file, data.index, meta_data_handler=meta_data_handler)
        gene_metadata = self.load_gene_metadata_tsv(gene_data_file, gene_name_column)

        data = data.transpose() if transpose_expression_data else data
        data = InferelatorData(data,
                               meta_data=meta_data,
                               gene_data=gene_metadata)

        # Make sure bytestrings are decoded
        _safe_dataframe_decoder(data.gene_data)
        _safe_dataframe_decoder(data.meta_data)

        return data

    def load_data_tenx(self, tenx_path, meta_data_file=None, meta_data_handler=DEFAULT_METADATA, gene_data_file=None,
                       gene_name_column=None):

        mtx_file, mtx_obs, mtx_feature = None, None, None

        for datafile in _TENX_MTX:
            if self._file_exists(self.filename_path_join(tenx_path, datafile)):
                mtx_file = self.filename_path_join(tenx_path, datafile)

        for datafile in _TENX_BARCODES:
            if self._file_exists(self.filename_path_join(tenx_path, datafile)):
                mtx_obs = self.filename_path_join(tenx_path, datafile)

        for datafile in _TENX_FEATURES:
            if self._file_exists(self.filename_path_join(tenx_path, datafile)):
                mtx_feature = self.filename_path_join(tenx_path, datafile)

        if mtx_file is None:
            msg = "Cannot find 10x files ({f}) in path ({p})".format(f=" or ".join(_TENX_MTX), p=tenx_path)
            raise FileNotFoundError(msg)
        else:
            return self.load_data_mtx(mtx_file, mtx_obs=mtx_obs, mtx_feature=mtx_feature, meta_data_file=meta_data_file,
                                      meta_data_handler=meta_data_handler, gene_data_file=gene_data_file,
                                      gene_name_column=gene_name_column)

    def load_data_tsv(self, expression_matrix_file, transpose_expression_data=False, meta_data_file=None,
                      meta_data_handler=DEFAULT_METADATA, expression_matrix_metadata=None, gene_data_file=None,
                      gene_name_column=None):

        Debug.vprint("Loading expression data file {file}".format(file=expression_matrix_file), level=0)

        # Load expression data
        data = self.input_dataframe(expression_matrix_file)
        if expression_matrix_metadata is not None:
            meta_cols = data.columns.intersection(expression_matrix_metadata)
            slice_meta_data = data.loc[:, meta_cols].copy()
            data = data.drop(meta_cols, axis=1)
        else:
            slice_meta_data = None

        if meta_data_file is None and slice_meta_data is not None:
            meta_data = None
        else:
            sample_labels = data.columns if transpose_expression_data else data.index
            meta_data = self.load_metadata_tsv(meta_data_file, sample_labels, meta_data_handler=meta_data_handler)

        meta_data = pd.concat((meta_data, slice_meta_data), axis=1)

        gene_metadata = self.load_gene_metadata_tsv(gene_data_file, gene_name_column)

        # Pack all data structures into an InferelatorData object
        data = InferelatorData(data,
                               transpose_expression=transpose_expression_data,
                               meta_data=meta_data,
                               gene_data=gene_metadata)

        self._check_loaded_data(data, filename=expression_matrix_file)
        return data

    def load_metadata_tsv(self, meta_data_file, sample_labels, meta_data_handler=None):
        # Load metadata
        meta_data_handler = MetadataHandler.get_handler(meta_data_handler)
        if meta_data_file is not None:
            Debug.vprint("Loading metadata file {file}".format(file=meta_data_file), level=0)
            meta_data = meta_data_handler.check_loaded_meta_data(self.input_dataframe(meta_data_file, index_col=None))
        else:
            Debug.vprint("No metadata provided. Creating a generic metadata", level=0)
            meta_data = meta_data_handler.create_default_meta_data(sample_labels)

        return meta_data

    def load_gene_metadata_tsv(self, gene_data_file, gene_name_column):
        # Load gene metadata
        if gene_data_file is None and gene_name_column is None:
            return None
        elif gene_data_file is None or gene_name_column is None:
            raise ValueError("Gene_metadata_file and gene_list_index must both be set if either is")

        Debug.vprint("Loading gene metadata from file {file}".format(file=gene_data_file), level=0)
        gene_metadata = self.input_dataframe(gene_data_file, index_col=None)

        # Validate that the gene_metadata can be properly read, if loaded
        if gene_name_column in gene_metadata:
            gene_metadata.index = gene_metadata[gene_name_column]
        else:
            msg = "Column {c} not found in gene data file [{h}]".format(c=gene_name_column,
                                                                        h=" ".join(gene_metadata.columns))
            raise ValueError(msg)

        return gene_metadata

    def input_dataframe(self, filename, **kwargs):
        """
        Read a file in as a pandas dataframe
        """
        Debug.vprint("Loading data file: {a}".format(a=self.input_path(filename)), level=2)

        # Use any kwargs for this function and any file settings from default
        if self._file_format_settings is not None and filename in self._file_format_settings:
            file_settings = self._file_format_settings[filename]
        else:
            file_settings = cp.copy(DEFAULT_PANDAS_TSV_SETTINGS)

        file_settings.update(kwargs)

        # Load a dataframe
        return pd.read_csv(self.input_path(filename), **file_settings)

    def input_path(self, filename):
        """
        Join filename to input_dir

        :param filename: Path to some file that needs to be attached to the input path
        :type filename: str
        :return: File joined to input_dir instance variable
        :rtype: str
        """

        return self.filename_path_join(self.input_dir, filename)

    @staticmethod
    def _file_exists(filename):
        return filename is not None and os.path.isfile(filename)

    @staticmethod
    def _load_list_from_file(filename):
        return pd.read_csv(filename, sep="\t", header=None)[0].tolist() if filename is not None else None

    @staticmethod
    def _check_loaded_data(data, filename=None):
        msg = "Loaded {f}:\n".format(f=filename) if filename is not None else ""

        nnf, non_finite_genes = data.non_finite
        if nnf > 0:
            msg += "\t{n} genes with non-finite expression ({g})\n".format(n=nnf, g=" ".join(non_finite_genes))

        msg += "Data loaded: {dt}".format(dt=str(data))
        Debug.vprint(msg, level=0)

    @staticmethod
    def filename_path_join(path, filename):
        """
        Join filename to path
        """

        # Raise an error if filename is None
        if filename is None:
            raise ValueError("Cannot create a path to a filename set as None")

        # Return an absolute path unchanged
        elif os.path.isabs(filename):
            return filename

        # If path is set, join filename to it and return that
        elif path is not None:
            return InferelatorDataLoader.make_path_safe(os.path.join(path, filename))

        # If path is not set, convert the filename to absolute and return it
        else:
            return InferelatorDataLoader.make_path_safe(filename)

    @staticmethod
    def make_path_safe(path):
        """
        Expand relative paths to absolute paths. Pass None through.
        :param path: str
        :return: str
        """
        if path is not None:
            return os.path.abspath(os.path.expanduser(path))
        else:
            return None


def _safe_dataframe_decoder(data_frame, encoding='utf-8'):
    """
    Decode dataframe bytestrings

    :param data_frame: pd.DataFrame
    """

    if _is_dtype_object(data_frame.index.dtype):
        data_frame.index = _decode_series(data_frame.index, encoding=encoding)

    if _is_dtype_object(data_frame.columns.dtype):
        data_frame.columns = _decode_series(data_frame.columns, encoding=encoding)

    for col in data_frame.columns:
        if _is_dtype_object(data_frame[col].dtype):
            data_frame[col] = _decode_series(data_frame[col], encoding=encoding)


def _is_dtype_object(dtype):
    if pat.is_object_dtype(dtype):
        return True
    elif pat.is_categorical_dtype(dtype):
        return pat.is_object_dtype(dtype.categories.dtype)
    else:
        return False


def _decode_series(series, encoding):
    """
    Decode and return a series or index object from pandas

    :param series: pd.Series, pd.Index
    :param encoding: str
    :return: pd.Series, pd.Index
    """

    if pat.is_categorical_dtype(series):
        series.cat.categories = _decode_series(series.dtype.categories, encoding=encoding)
        return series

    _new_series = series.str.decode(encoding).values
    _no_decode = pd.isna(_new_series)

    if np.all(_no_decode):
        return series
    
    _new_series[_no_decode] = series.values[_no_decode]

    try:
        new_series = pd.Series(_new_series, index=series.index)
    except AttributeError:
        new_series = pd.Index(_new_series)

    return new_series



import os
from setuptools import setup, find_packages

# Current Inferelator Version Number
version = "0.5.3"


# Description from README.md
base_dir = os.path.dirname(os.path.abspath(__file__))
long_description = "\n\n".join([open(os.path.join(base_dir, "README.md"), "r").read()])

# Required packages
requires = [
    "numpy >=1.14.0",
    "scipy >=0.9",
    "pandas >=0.24.0",
    "scikit-learn",
    "matplotlib >=1.5.1",
    "anndata >=0.7.4"
]

setup(
    name="inferelator",
    version=version,
    description="Inferelator: Network Inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flatironinstitute/inferelator",
    project_urls={
        "Documentation": "https://inferelator.readthedocs.io/",
    },
    author="Chris Jackson",
    author_email="cj59@nyu.edu",
    maintainer="Chris Jackson",
    maintainer_email="cj59@nyu.edu",
    packages=find_packages(include=["inferelator", "inferelator.*"], exclude=["tests", "*.tests"]),
    zip_safe=False,
    install_requires=requires,
    python_requires=">=3.5",
    tests_require=["coverage", "nose", "bio-test-artifacts", "tables"],
    test_suite="nose.collector",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta"
    ]
)



"""
This is a central location for dataframe column names used in the postprocessing modules

They're weirdly named for historical reasons and can be changed with no consequence unless you have other
code that uses these names
"""

BETA_SIGN_COLUMN = "beta.sign.sum"
MEDIAN_EXPLAIN_VAR_COLUMN = "var.exp.median"
CONFIDENCE_COLUMN = "combined_confidences"
BETA_THRESHOLD_COLUMN = "beta_threshold"
GOLD_STANDARD_COLUMN = "gold_standard"
PRIOR_COLUMN = "prior"
TARGET_COLUMN = "target"
REGULATOR_COLUMN = "regulator"

# Precision/Recall

PRECISION_COLUMN = "precision"
RECALL_COLUMN = "recall"

# Matthew Correlation Coef

MCC_COLUMN = "MCC"

# F1 Score

F1_COLUMN = "F1"

# Confusion Matrix

TP, FP, TN, FN = 'TP', 'FP', 'TN', 'FN'




"""
This is a manager which will take an Inferelator workflow and repeatedly run it with different parameters.
This is implemented using deep copies; it is therefore memory-intensive.
"""

from __future__ import print_function

# I hate py2 now
try:
    from builtins import FileExistsError
except ImportError:
    FileExistsError = OSError

import os
import warnings
import csv
import types
import copy
import itertools
import numpy as np
import pandas as pd


from inferelator.distributed.inferelator_mp import MPControl
from inferelator.utils import Validator as check
from inferelator import utils
from inferelator import workflow
from inferelator.postprocessing.model_metrics import MetricHandler


class CrossValidationManager(object):
    """
    Crossvalidate an Inferelator Workflow
    """

    # Output settings
    _csv_writer = None  # csv.csvwriter
    _csv_header = None  # list[]
    output_file_name = "crossvalidation_performance.tsv"  # str
    _csv_writer_object = csv.writer
    _csv_file_handle = None

    # Grid search parameters
    grid_params = None
    grid_param_values = None

    # Dropin/dropout categories
    dropout_column = None
    dropout_max_size = None
    dropout_seed = None

    dropin_column = None
    dropin_max_size = None
    dropin_seed = None

    # Size subsampling categories
    size_sample_vector = None
    size_sample_stratified_column = None
    size_sample_with_replacement = False
    size_sample_seed = None

    # Workflow storage
    _baseline_workflow = None
    _baseline_output_dir = None
    _baseline_input_dir = None

    @property
    def output_dir(self):
        if self._baseline_workflow is None:
            return self._baseline_output_dir
        else:
            return self._baseline_workflow.output_dir

    @output_dir.setter
    def output_dir(self, path):
        if self._baseline_workflow is not None:
            self._baseline_workflow.output_dir = path
        else:
            self._baseline_output_dir = path

    @property
    def input_dir(self):
        if self._baseline_workflow is None:
            return self._baseline_input_dir
        else:
            return self._baseline_workflow.input_dir

    @input_dir.setter
    def input_dir(self, path):
        if self._baseline_workflow is not None:
            self._baseline_workflow.input_dir = path
        else:
            self._baseline_input_dir = path

    @property
    def workflow(self):
        if self._baseline_workflow is None:
            raise AttributeError("No workflow has been provided")
        return self._baseline_workflow

    @workflow.setter
    def workflow(self, wkf):
        assert check.argument_is_subclass(wkf, workflow.WorkflowBase, allow_none=True)
        if self._baseline_workflow is not None:
            warnings.warn("Replacing stored workflow with a new workflow")
        self._baseline_workflow = wkf

    def __init__(self, workflow_object=None):
        """
        Create a new CrossValidationManager instance and give it a workflow

        :param workflow_object: The workflow to run crossvalidation with
        :type workflow_object: Workflow

        """
        self.workflow = workflow_object

    def add_gridsearch_parameter(self, param_name, param_vector):
        """
        Set a parameter to search through by exhaustive grid search

        :param param_name: The workflow parameter to change for each run
        :type param_name: str
        :param param_vector: An iterable with values to use for the parameter
        :type param_vector: iterable

        """

        if self.grid_param_values is None:
            self.grid_param_values = {}

        if self.grid_params is None:
            self.grid_params = []

        self.grid_params.append(param_name)

        if utils.is_string(param_vector):
            self.grid_param_values[param_name] = [param_vector]
        else:
            try:
                # There's probably a better way to check and see if something is iterable but I don't care
                [True for _ in param_vector]
                self.grid_param_values[param_name] = param_vector
            except TypeError:
                self.grid_param_values[param_name] = [param_vector]

    def add_grouping_dropout(self, metadata_column_name, group_size=None, seed=42):
        """
        Drop each group (defined by a metadata column) and run modeling on all of the other groups.

        :param metadata_column_name: Metadata column which has different values for each group
        :type metadata_column_name: str
        :param group_size: The maximum size of each group. Groups will be downsampled to the same size if this is not
            set to None. Default is None.
        :type group_size: int, None
        :param seed: The random seed to use for the group downsampling
            (this is not the same as the seed passed to the workflow)
        :type seed: int
        """

        self.dropout_column = metadata_column_name
        self.dropout_max_size = group_size
        self.dropout_seed = seed

    def add_grouping_dropin(self, metadata_column_name, group_size=None, seed=42):
        """
        Run modeling on each group (defined by a metadata column) individually.

        :param metadata_column_name: Metadata column which has different values for each group
        :type metadata_column_name: str
        :param group_size: The maximum size of each group. Groups will be downsampled to the same size if this is not
            set to None. Default is None.
        :type group_size: int, None
        :param seed: The random seed to use for the group downsampling
            (this is not the same as the seed passed to the workflow)
        :type seed: int
        """

        self.dropin_column = metadata_column_name
        self.dropin_max_size = group_size
        self.dropin_seed = seed

    def add_size_subsampling(self, size_vector, stratified_column_name=None, with_replacement=False, seed=42):
        """
        Resample expression data to a ratio of the original data.

        :param size_vector: An iterable with numeric ratios for downsampling. These values must be between 0 and 1.
        :type size_vector: iterable(floats)
        :param stratified_column_name: Set this to stratify sampling (to maintain group size ratios). If None, do not
            maintain group size ratios. Default is None.
        :type stratified_column_name: str, None
        :param with_replacement: Do sampling with or without replacement. Defaults to False
        :type with_replacement: bool
        :param seed: The random seed to use when selecting observations
            (this is not the same as the seed passed to the workflow)
        :param seed: int
        """

        try:
            [check.argument_numeric(val, low=0, high=1) for val in size_vector]
        except ValueError as err:
            utils.Debug.vprint("Size sampling parameter error: {err}".format(err=str(err)), level=0)
            raise

        self.size_sample_vector = size_vector
        self.size_sample_stratified_column = stratified_column_name
        self.size_sample_with_replacement = with_replacement
        self.size_sample_seed = seed

    def run(self):
        """
        Execute crossvalidation after all setup
        """

        # Create output path
        self._harmonize_paths()
        # Open a CSV file handle
        self._create_writer()
        # Load and check data into a workflow object
        self._initial_data_load()
        self._check_metadata()
        self._check_grid_search_params_exist()

        # Run base grid search
        results = self._grid_search()

        # Run size sampling
        if self.size_sample_vector is not None:
            results.extend(self._size_cv())

        # Run dropin
        if self.dropin_column is not None:
            results.extend(self._dropin_cv())

        # Run dropout
        if self.dropout_column is not None:
            results.extend(self._dropout_cv())

        self._destroy_writer()

        return results

    def append_to_path(self, var_name, to_append):
        """
        Add a string to an existing path variable in class
        """

        # Get the path
        path = getattr(self, var_name)

        # If neither are set just die
        if path is None:
            raise ValueError("Cannot append {to_append} to {var_name} (Which is None)".format(to_append=to_append,
                                                                                              var_name=var_name))
        setattr(self, var_name, os.path.join(path, to_append))

    def _create_writer(self):
        """
        Create a CSVWriter and stash it in self.writer
        """

        if MPControl.is_master:
            # Create a CSV header from grid search param names
            self._csv_header = copy.copy(self.grid_params) if self.grid_params is not None else []

            # Add Test & Value columns for dropouts/etc
            self._csv_header.extend(["Test", "Value", "Num_Obs"])

            # Also add the metric name
            self._csv_header.extend(MetricHandler.get_metric(self.workflow.metric).all_names())

            # Create a CSV writer
            self._create_output_path()
            self._open_csv_handle()

            self._csv_writer = self._csv_writer_object(self._csv_file_handle,
                                                       delimiter="\t", lineterminator="\n", quoting=csv.QUOTE_NONE)

            # Write the header line
            self._csv_writer.writerow(self._csv_header)

    def _destroy_writer(self):
        """
        Delete the CSVWriter and close the file handle
        """
        if MPControl.is_master:
            self._csv_file_handle.close()
            self._csv_writer = None

    def _harmonize_paths(self):
        """
        If _baseline is set, copy it to the workflow
        If _baseline is not set, copy to it from the workflow
        """
        if self._baseline_output_dir is None and self.workflow.output_dir is None:
            raise ValueError("No output path has been provided to either crossvalidation or workflow")
        elif self._baseline_output_dir is None:
            self._baseline_output_dir = self.workflow.output_dir
        elif self.workflow.output_dir is None:
            self.workflow.output_dir = self._baseline_output_dir

        try:
            check.argument_subpath(self.workflow.output_dir, self._baseline_output_dir)
        except ValueError:
            warnings.warn("Workflow output path is {p}; resetting to {a}".format(a=self.workflow.output_dir,
                                                                                 p=self._baseline_output_dir))
            self.workflow.output_dir = self._baseline_output_dir

        if self._baseline_input_dir is None and self.workflow.input_dir is None:
            raise ValueError("No input path has been provided to either crossvalidation or workflow")

        if self._baseline_input_dir is None:
            self._baseline_input_dir = self.workflow.input_dir
        if self.workflow.input_dir is None:
            self.workflow.input_dir = self._baseline_input_dir

    def _open_csv_handle(self):
        """
        Open and return a file handle to the CSV output file
        """
        csv_file_name = os.path.join(self.output_dir, self.output_file_name)
        self._csv_file_handle = open(csv_file_name, mode="w", buffering=1)

    def _create_output_path(self):
        """
        Create the output path
        """
        self.output_dir = os.path.abspath(os.path.expanduser(self.output_dir))

        try:
            os.makedirs(self.output_dir)
        except FileExistsError:
            pass

    def _initial_data_load(self):
        """
        Load data into the workflow
        """
        # Load data with the baseline get_data() function
        self.workflow.get_data()

        # Blow up the get_data() function so that it doesn't get re-run
        def mock_get_data(slf):
            pass

        self.workflow.get_data = types.MethodType(mock_get_data, self.workflow)

    def _get_workflow_copy(self):
        """
        Copies and returns the workflow which has loaded data
        """

        return copy.deepcopy(self.workflow)

    def _grid_search(self, test=None, value=None, mask_function=None):
        """

        :param test: str
        :param value: str
        :param mask_function: func
            A function which produces a pd.Series(bool) [N] mask for selecting observations
        :return:
        """

        if test is not None:
            utils.Debug.vprint("Grid search for  {t} [{v}]".format(t=test, v=value))

        # This is unpacked in the same order that is used in the header
        ordered_unpack = [self.grid_param_values[param] for param in self.grid_params]

        results = []

        for param_values in itertools.product(*ordered_unpack):
            params = zip(self.grid_params, param_values)
            csv_line = []
            output_path = []

            # Get the workflow and set the CV parameters
            cv_workflow = self._get_workflow_copy()
            cv_workflow.create_output_dir()

            # Drop any observations which are False in the mask (if set)
            if mask_function is not None:
                mask = mask_function()
                cv_workflow.expression_matrix.drop(cv_workflow.expression_matrix.columns[~mask], axis=1, inplace=True)
                cv_workflow.meta_data.drop(cv_workflow.meta_data.index[~mask], axis=0, inplace=True)
                n_obs = mask.sum()
            else:
                n_obs = cv_workflow._num_obs

            for name, param_value in params:
                csv_line.append(param_value)
                output_path.append(str(name) + "_" + str(param_value))
                setattr(cv_workflow, name, param_value)
                utils.Debug.vprint("Setting crossvalidation param {p} to {v}".format(p=name, v=param_value), level=0)

            # Set the parameters into the output path
            if test is not None:
                output_path = "_".join(map(str, output_path + [test, value]))

            else:
                output_path = "_".join(map(str, output_path))

            # Run the workflow
            cv_workflow.append_to_path("output_dir", output_path)
            utils.Debug.vprint("Writing results to {p} ".format(p=cv_workflow.output_dir), level=1)
            result = cv_workflow.run()

            # Output the results to the CSV file
            csv_line.extend([test, value, n_obs] + [result.all_scores[n] for n in result.all_names])

            results.append(((test, value), result))

            if MPControl.is_master:
                self._csv_writer.writerow(csv_line)

            del cv_workflow

        return results

    def _check_grid_search_params_exist(self):
        """
        Determine if the parameters for grid search are workflow parameters
        """

        for param in self.grid_params:
            if hasattr(self.workflow, param):
                pass
            else:
                raise ValueError("Parameter {p} for GridCV does not appear to be a valid parameter".format(p=param))

    def _check_metadata(self):
        """
        Make sure that any set metadata columns exist after loading
        """

        if self.dropout_column is not None:
            self._check_metadata_column_exists(self.dropout_column)
        if self.dropin_column is not None:
            self._check_metadata_column_exists(self.dropin_column)
        if self.size_sample_stratified_column is not None:
            self._check_metadata_column_exists(self.size_sample_stratified_column)

    def _check_metadata_column_exists(self, col_name):
        """
        Check to make sure the metadata column exists

        :param col_name: str
        """

        if col_name in self.workflow.meta_data.columns:
            return True
        else:
            raise ValueError("Column {col} is not present in the loaded metadata".format(col=col_name))

    def _dropout_cv(self):
        """
        Run grid search on all data minus one group at a time
        """

        meta_data = self.workflow.meta_data.copy()
        col = self.dropout_column
        max_size = self.dropout_max_size

        unique_groups = meta_data[col].unique().tolist()

        results = []

        # Downsample if max_size is set for a comparable all-group control
        if len(unique_groups) > 2 and max_size is not None:
            rgen = np.random.RandomState(self.dropout_seed - 1)

            def mask_function():

                include_mask = pd.Series(False, index=meta_data.index)

                for g in unique_groups:
                    include_mask = include_mask | group_index(meta_data, col, g, rgen=rgen, max_size=max_size)

                return include_mask

            result = self._grid_search(test="dropout", value="all", mask_function=mask_function)
            results.append((("dropout", "all"), result))

        # Iterate through groups and drop one
        for i, group in enumerate(unique_groups):
            rgen = np.random.RandomState(self.dropout_seed + i)

            def mask_function():
                if max_size is None:
                    return meta_data[col] != group
                else:
                    include_mask = pd.Series(False, index=meta_data.index)

                    # For each factor in the stratified column
                    for g in unique_groups:
                        if g == group:
                            continue

                        include_mask = include_mask | group_index(meta_data, col, g, rgen=rgen, max_size=max_size)

                    return include_mask

            result = self._grid_search(test="dropout", value=group, mask_function=mask_function)
            results.append((("dropout", group), result))

        return results

    def _dropin_cv(self):
        """
        Run grid search on one group from the data at a time
        """

        meta_data = self.workflow.meta_data.copy()
        col = self.dropin_column
        max_size = self.dropin_max_size

        unique_groups = meta_data[col].unique().tolist()

        results = []

        if len(unique_groups) > 1 and max_size:
            rgen = np.random.RandomState(self.dropin_seed - 1)

            def mask_function():
                return group_index(meta_data, col, group=None, rgen=rgen, max_size=max_size)

            result = self._grid_search(test="dropin", value="all", mask_function=mask_function)
            results.append((("dropin", "all"), result))

        for i, group in enumerate(unique_groups):
            rgen = np.random.RandomState(self.dropin_seed + i)

            def mask_function():
                if max_size is None:
                    return meta_data[col] == group
                else:
                    return group_index(meta_data, col, group, rgen=rgen, max_size=max_size)

            result = self._grid_search(test="dropin", value=group, mask_function=mask_function)
            results.append((("dropin", group), result))

        return results

    def _size_cv(self):
        """
        Run grid search on a subset of the data
        """

        results = []

        for i, size_ratio in enumerate(self.size_sample_vector):
            rgen = np.random.RandomState(self.size_sample_seed + i)
            meta_data = self.workflow.meta_data.copy()

            if self.size_sample_stratified_column is not None:
                strat_col = self.size_sample_stratified_column

                def data_masker():
                    unique_groups = meta_data[strat_col].unique().tolist()
                    data_mask = pd.Series(False, index=meta_data.index)

                    # For each factor in the stratified column
                    for group in unique_groups:
                        data_mask = data_mask | group_index(meta_data, strat_col, group, size_ratio=size_ratio,
                                                            rgen=rgen)

                    return data_mask

            else:

                def data_masker():
                    n_obs = meta_data.shape[0]
                    size = max(int(n_obs * size_ratio), 1)
                    size_mask = [True] * size + [False] * (n_obs - size)
                    rgen.shuffle(size_mask)
                    return pd.Series(size_mask, index=meta_data.index)

            result = self._grid_search(test="size", value=str(size_ratio), mask_function=data_masker)
            results.append((("size", size_ratio), result))

        return results


def group_index(meta, meta_col, group=None, size_ratio=None, rgen=None, max_size=None):
    rgen = rgen if rgen is not None else np.random

    # Find the observations in this group
    if group is None:
        group_mask = pd.Series(True, index=meta.index)
    else:
        group_mask = meta[meta_col] == group

    n_group = group_mask.sum()

    if n_group == 0:
        return group_mask

    # Decide how big to make the group (size * size_ratio)
    if size_ratio is not None:
        size = max(int(n_group * size_ratio), 1)
    else:
            size = max(n_group, 1)

    # Set max_size if that argument has been provided
    if max_size is not None:
        size = min(size, max_size)

    if size != n_group:
        size_mask = [True] * size + [False] * (n_group - size)
        rgen.shuffle(size_mask)
        group_mask.loc[group_mask] = size_mask

    return group_mask



from inferelator.preprocessing.tfa import TFA
from inferelator.utils import InferelatorData


class VelocityTFA(TFA):

    def compute_transcription_factor_activity(self, prior, expression_data, expression_data_halftau=None,
                                              keep_self=False, tau=None):

        prior, activity_tfs, expr_tfs = self._check_prior(prior, expression_data, keep_self=keep_self)

        if len(activity_tfs) > 0:
            activity = self._calculate_activity(prior.loc[:, activity_tfs].values, expression_data)
        else:
            raise ValueError("TFA cannot be calculated; prior matrix has no edges")

        return InferelatorData(activity, gene_names=activity_tfs, sample_names=expression_data.sample_names,
                               meta_data=expression_data.meta_data)





from __future__ import print_function, unicode_literals, division

import copy as cp
import gc
import math
import pandas as pd
import numpy as np
import scipy.sparse as sparse
import scipy.stats
import pandas.api.types as pat
from sklearn.preprocessing import StandardScaler
import scipy.io
from anndata import AnnData
from inferelator.utils import Debug, Validator


def dot_product(a, b, dense=True, cast=True):
    """
    Dot product two matrices together. Allow either matrix (or both or neither) to be sparse.

    :param a:
    :param b:
    :param dense:
    :param cast:
    :return:
    """
    if sparse.isspmatrix(a) and sparse.isspmatrix(b):
        return a.dot(b).A if dense else a.dot(b)
    elif sparse.isspmatrix(a):
        return a.dot(sparse.csr_matrix(b)).A if dense else a.dot(sparse.csr_matrix(b))
    elif sparse.isspmatrix(b):
        return np.dot(a, b.A)
    else:
        return np.dot(a, b)


class DotProduct:

    _dot_func = dot_product

    @classmethod
    def set_mkl(cls, mkl=True):

        # If the MKL flag is None, don't change anything
        if mkl is None:
            pass

        # If the MKL flag is True, use the dot_product_mkl function when .dot() is called
        if mkl:
            try:
                from sparse_dot_mkl import get_version_string, dot_product_mkl as dp
                msg = "Matrix multiplication will use sparse_dot_mkl package with MKL: {m}"
                vstring = get_version_string()
                Debug.vprint(msg.format(m=vstring if vstring is not None else "Install mkl-service for details"),
                             level=2)

                cls._dot_func = dp

            # If it isn't available, use the scipy/numpy functions instead
            except ImportError as err:
                Debug.vprint("Unable to load MKL with sparse_dot_mkl:\n" + str(err), level=0)
                cls._dot_func = dot_product

        # If the MKL flag is True, use the python (numpy/scipy) functions when .dot() is called
        else:
            Debug.vprint("Matrix multiplication will use Numpy; this is not advised for sparse data", level=2)
            cls._dot_func = dot_product

    @classmethod
    def dot(cls, *args, **kwargs):
        return cls._dot_func(*args, **kwargs)


def df_from_tsv(file_like, has_index=True):
    "Read a tsv file or buffer with headers and row ids into a pandas dataframe."
    return pd.read_csv(file_like, sep="\t", header=0, index_col=0 if has_index else False)


def df_set_diag(df, val, copy=True):
    """
    Sets the diagonal of a dataframe to a value. Diagonal in this case is anything where row label == column label.

    :param df: pd.DataFrame
        DataFrame to modify
    :param val: numeric
        Value to insert into any cells where row label == column label
    :param copy: bool
        Force-copy the dataframe instead of modifying in place
    :return: pd.DataFrame / int
        Return either the modified dataframe (if copied) or the number of cells modified (if changed in-place)
    """

    # Find all the labels that are shared between rows and columns
    isect = df.index.intersection(df.columns)

    if copy:
        df = df.copy()

    # Set the value where row and column names are the same
    for i in range(len(isect)):
        df.loc[isect[i], isect[i]] = val

    if copy:
        return df
    else:
        return len(isect)


def array_set_diag(arr, val, row_labels, col_labels):
    """
    Sets the diagonal of an 2D array to a value. Diagonal in this case is anything where row label == column label.

    :param arr: Array to modify in place
    :type arr: np.ndarray
    :param val: Value to insert into any cells where row label == column label
    :type val: numeric
    :param row_labels: Labels which correspond to the rows
    :type row_labels: list, pd.Index
    :param col_labels: Labels which correspond to the columns
    :type col_labels: list, pd.Index
    :return: Return the number of common row and column labels
    :rtype: int
    """

    if arr.ndim != 2:
        raise ValueError("Array must be 2D")

    # Find all the labels that are shared between rows and columns
    isect = set(row_labels).intersection(col_labels)

    # Set the value where row and column names are the same
    for i in isect:
        arr[row_labels == i, col_labels == i] = val

    return len(isect)


def make_array_2d(arr):
    """
    Changes array shape from 1d to 2d if needed (in-place)
    :param arr:  np.ndarray
    """
    if arr.ndim == 1:
        arr.shape = (arr.shape[0], 1)


def melt_and_reindex_dataframe(data_frame, value_name, idx_name="target", col_name="regulator"):
    """
    Take a pandas dataframe and melt it into a one column dataframe (with the column `value_name`) and a multiindex
    of the original index + column
    :param data_frame: pd.DataFrame [M x N]
        Meltable dataframe
    :param value_name: str
        The column name for the values of the dataframe
    :param idx_name: str
        The name to assign to the original data_frame index values
    :param col_name: str
        The name to assign to the original data_frame column values
    :return: pd.DataFrame [(M*N) x 1]
        Melted dataframe with a single column of values and a multiindex that is the original index + column for
        that value
    """

    # Copy the dataframe and move the index to a column
    data_frame = data_frame.copy()
    data_frame[idx_name] = data_frame.index

    # Melt it into a [(M*N) x 3] dataframe
    data_frame = data_frame.melt(id_vars=idx_name, var_name=col_name, value_name=value_name)

    # Create a multiindex and then drop the columns that are now in the index
    data_frame.index = pd.MultiIndex.from_frame(data_frame.loc[:, [idx_name, col_name]])
    del data_frame[idx_name]
    del data_frame[col_name]

    return data_frame


def scale_vector(vec, ddof=1):
    """
    Take a vector and normalize it to a mean 0 and standard deviation 1 (z-score)

    :param vec: A 1d vector to be normalized
    :type vec: np.ndarray, sp.sparse.spmatrix
    :param ddof: The delta degrees of freedom for variance calculation
    :type ddof: int
    :return: A centered and scaled vector
    :rtype: np.ndarray
    """

    # Convert a sparse vector to a dense vector
    if sparse.isspmatrix(vec):
        vec = vec.A

    # Return 0s if the variance is 0
    if np.var(vec) == 0:
        return np.zeros(vec.shape, dtype=float)

    # Otherwise scale with scipy.stats.zscore
    else:
        return scipy.stats.zscore(vec, axis=None, ddof=ddof)


def apply_window_vector(vec, window, func):
    """
    Apply a function to a 1d array by windows.
    For logical testing of an array without having to allocate a full array of bools

    :param vec: A 1d vector to be normalized
    :type vec: np.ndarray, sp.sparse.spmatrix
    :param window: The window size to process
    :type window: int
    :param func: A function that produces an aggregate result for the window
    :type func: callable
    :return:
    """
    steps = math.ceil(len(vec) / window)
    return np.array([func(vec[i * window:min((i + 1) * window, len(vec))]) for i in range(steps)])


class InferelatorData(object):
    """ Store inferelator data in an AnnData object. This will always be Samples by Genes """

    name = None

    _adata = None

    @property
    def _is_integer(self):
        return pat.is_integer_dtype(self._adata.X.dtype)

    @property
    def expression_data(self):
        return self._adata.X

    @expression_data.setter
    def expression_data(self, new_data):
        self._adata.X = new_data

    @property
    def values(self):
        return self._adata.X

    @property
    def _data(self):
        if self.is_sparse:
            return self._adata.X.data
        else:
            return self._adata.X

    @_data.setter
    def _data(self, new_data):
        if self.is_sparse:
            self._adata.X.data = new_data
        else:
            self._adata.X = new_data

    @property
    def _data_mem_usage(self):
        if self.is_sparse:
            return self._adata.X.data.nbytes + self._adata.X.indices.nbytes + self._adata.X.indptr.nbytes
        else:
            return self._adata.X.nbytes

    @property
    def meta_data(self):
        return self._adata.obs

    @meta_data.setter
    def meta_data(self, new_meta_data):

        if isinstance(new_meta_data, InferelatorData):
            new_meta_data = new_meta_data.meta_data

        # Reindex the new metadata to match the existing sample names
        new_meta_data = new_meta_data.copy()
        new_meta_data.index = new_meta_data.index.astype(str)

        # Force unique names by appending values
        if self._adata.obs_names.nunique() != self.num_obs:
            self._adata.obs_names_make_unique()

        # Drop duplicate names on the new meta data
        if new_meta_data.index.nunique() != new_meta_data.shape[0]:
            new_meta_data = new_meta_data.loc[~new_meta_data.duplicated(), :]

        # If the new one is the right size, force it in one way or the other
        # Reindex the metadata to match the sample names
        try:
            new_meta_data = new_meta_data.reindex(self.sample_names)
        except ValueError:
            # If the metadata is the wrong size, angrily die
            if new_meta_data.shape[0] != self.num_obs:
                msg = "Metadata size {sh1} does not match data ({sh2})".format(sh1=new_meta_data.shape,
                                                                                          sh2=self.num_obs)
                raise ValueError(msg)
            new_meta_data.index = self.sample_names

        if len(self._adata.obs.columns) > 0:
            keep_columns = self._adata.obs.columns.difference(new_meta_data.columns)
            self._adata.obs = pd.concat((new_meta_data, self._adata.obs.loc[:, keep_columns]), axis=1)
        else:
            self._adata.obs = new_meta_data

    @property
    def gene_data(self):
        return self._adata.var

    @gene_data.setter
    def gene_data(self, new_gene_data):

        if isinstance(new_gene_data, InferelatorData):
            new_gene_data = new_gene_data.gene_data

        new_gene_data = new_gene_data.copy()
        new_gene_data.index = new_gene_data.index.astype(str)
        # Use the intersection of this and the expression data genes to make a list of gene names to keep
        self._adata.uns["trim_gene_list"] = new_gene_data.index.intersection(self._adata.var.index)

        new_gene_data = new_gene_data.reindex(self._adata.var_names)

        # Join any new columns to any existing columns
        # Update (overwrite) any columns in the existing meta data if they are in the new meta data
        if len(self._adata.var.columns) > 0:
            keep_columns = self._adata.var.columns.difference(new_gene_data.columns)
            self._adata.var = pd.concat((new_gene_data, self._adata.var.loc[:, keep_columns]), axis=1)
        else:
            self._adata.var = new_gene_data

    @property
    def gene_names(self):
        return self._adata.var_names

    @property
    def gene_counts(self):
        return self._adata.X.sum(axis=0).A.flatten() if self.is_sparse else self._adata.X.sum(axis=0)

    @property
    def gene_means(self):
        return self._adata.X.mean(axis=0).A.flatten() if self.is_sparse else self._adata.X.mean(axis=0)

    @property
    def gene_stdev(self):
        if self.is_sparse:
            return np.sqrt(StandardScaler(copy=False, with_mean=False).fit(self._adata.X).var_)
        else:
            return self._adata.X.std(axis=0, ddof=1)

    @property
    def sample_names(self):
        return self._adata.obs_names

    @property
    def sample_counts(self):
        return self._adata.X.sum(axis=1).A.flatten() if self.is_sparse else self._adata.X.sum(axis=1)

    @property
    def sample_means(self):
        return self._adata.X.mean(axis=1).A.flatten() if self.is_sparse else self._adata.X.mean(axis=1)

    @property
    def sample_stdev(self):
        if self.is_sparse:
            return np.sqrt(StandardScaler(copy=False, with_mean=False).fit(self._adata.X.T).var_)
        else:
            return self._adata.X.std(axis=1, ddof=1)

    @property
    def non_finite(self):
        if min(self._data.shape) == 0:
            return 0, None
        elif self.is_sparse:
            nnf = np.sum(apply_window_vector(self._adata.X.data, 1000000, lambda x: np.sum(~np.isfinite(x))))
            return nnf, ["GENES_NOT_ID_SPARSE_MATRIX"] if nnf > 0 else None
        else:
            non_finite = np.apply_along_axis(lambda x: np.sum(~np.isfinite(x)) > 0, 0, self._data)
            nnf = np.sum(non_finite)
            return nnf, self.gene_names[non_finite] if nnf > 0 else None

    @property
    def is_sparse(self):
        return sparse.issparse(self._adata.X)

    @property
    def shape(self):
        return self._adata.shape

    @property
    def num_obs(self):
        return self._adata.shape[0]

    @property
    def num_genes(self):
        return self._adata.shape[1]

    @property
    def uns(self):
        return self._adata.uns

    def __str__(self):
        msg = "InferelatorData [{dt} {sh}, Metadata {me}] Memory: {mem:.2f} MB"
        return msg.format(sh=self.shape, dt=self._data.dtype, me=self.meta_data.shape,
                          mem=(self._data_mem_usage / 1e6))

    def __init__(self, expression_data=None, transpose_expression=False, meta_data=None, gene_data=None,
                 gene_data_idx_column=None, gene_names=None, sample_names=None, dtype=None, name=None):
        """
        Create a new InferelatorData object

        :param expression_data: A tabular observations x variables matrix
        :type expression_data: np.array, scipy.sparse.spmatrix, anndata.AnnData, pd.DataFrame
        :param transpose_expression: Should the table be transposed. Defaults to False
        :type transpose_expression: bool
        :param meta_data: Meta data for observations. Needs to align to the expression data
        :type meta_data: pd.DataFrame
        :param gene_data: Meta data for variables.
        :type gene_data: pd.DataFrame
        :param gene_data_idx_column: The gene_data column which should be used to align to expression_data
        :type gene_data_idx_column: bool
        :param gene_names: Names to be used for variables.
            Will be inferred from a dataframe or anndata object if not set.
        :type gene_names: list, pd.Index
        :param sample_names: Names to be used for observations
            Will be inferred from a dataframe or anndata object if not set.
        :type sample_names: list, pd.Index
        :param dtype: Explicitly convert the data to this dtype if set.
            Only applies to data loaded from a pandas dataframe. Numpy arrays and scipy matrices will use existing type.
        :type dtype: np.dtype
        :param name: Name of this data structure.
        :type name: None, str
        """

        if expression_data is not None and isinstance(expression_data, pd.DataFrame):
            object_cols = expression_data.dtypes == object

            if sum(object_cols) > 0:
                object_data = expression_data.loc[:, object_cols]
                meta_data = object_data if meta_data is None else pd.concat((meta_data, object_data), axis=1)
                expression_data.drop(expression_data.columns[object_cols], inplace=True, axis=1)

            if dtype is None and all(map(lambda x: pat.is_integer_dtype(x), expression_data.dtypes)):
                dtype = 'int32'
            elif dtype is None:
                dtype = 'float64'

            self._make_idx_str(expression_data)

            if transpose_expression:
                self._adata = AnnData(X=expression_data.T, dtype=dtype)
            else:
                self._adata = AnnData(X=expression_data, dtype=dtype)

        elif expression_data is not None and isinstance(expression_data, AnnData):
            self._adata = expression_data

        elif expression_data is not None:
            self._adata = AnnData(X=expression_data.T if transpose_expression else expression_data,
                                  dtype=expression_data.dtype)

        else:
            self._adata = AnnData()

        if gene_names is not None and len(gene_names) > 0:
            self._adata.var_names = gene_names

        if sample_names is not None and len(sample_names) > 0:
            self._adata.obs_names = sample_names

        if meta_data is not None:
            self._make_idx_str(meta_data)
            self.meta_data = meta_data

        if gene_data is not None:
            if gene_data_idx_column is not None and gene_data_idx_column in gene_data:
                gene_data.index = gene_data[gene_data_idx_column]
            elif gene_data_idx_column is not None:
                msg = "No gene_data column {c} in {a}".format(c=gene_data_idx_column, a=" ".join(gene_data.columns))
                raise ValueError(msg)
            self._make_idx_str(gene_data)
            self.gene_data = gene_data

        self._cached = {}
        self.name = name

    def convert_to_float(self):
        """
        Convert the data in-place to a float dtype
        """

        if pat.is_float_dtype(self._data.dtype):
            return None
        elif self._data.dtype == np.int32:
            dtype = np.float32
        elif self._data.dtype == np.int64:
            dtype = np.float64
        else:
            raise ValueError("Data is not float, int32, or int64")

        # Create a new memoryview with a specific dtype
        float_view = self._data.view(dtype)

        # Assign the old data through the memoryview
        float_view[:] = self._data

        # Replace the old data with the newly converted data
        self._data = float_view

    def trim_genes(self, remove_constant_genes=True, trim_gene_list=None):
        """
        Remove genes (columns) that are unwanted from the data set. Do this in-place.

        :param remove_constant_genes:
        :type remove_constant_genes: bool
        :param trim_gene_list: This is a list of genes to KEEP.
        :type trim_gene_list: list, pd.Series, pd.Index
        """

        keep_column_bool = np.ones((len(self._adata.var_names),), dtype=bool)

        if trim_gene_list is not None:
            keep_column_bool &= self._adata.var_names.isin(trim_gene_list)
        if "trim_gene_list" in self._adata.uns:
            keep_column_bool &= self._adata.var_names.isin(self._adata.uns["trim_gene_list"])

        list_trim = len(self._adata.var_names) - np.sum(keep_column_bool)
        comp = 0 if self._is_integer else np.finfo(self.values.dtype).eps * 10

        if remove_constant_genes:
            nz_var = self.values.max(axis=0) - self.values.min(axis=0)
            nz_var = nz_var.A.flatten() if self.is_sparse else nz_var

            if np.any(np.isnan(nz_var)):
                raise ValueError("NaN values are present in the expression matrix; unable to remove var=0 genes")

            nz_var = comp < nz_var

            keep_column_bool &= nz_var
            var_zero_trim = np.sum(nz_var)
        else:
            var_zero_trim = 0

        if np.sum(keep_column_bool) == 0:
            err_msg = "No genes remain after trimming. ({lst} removed to match list, {v} removed for var=0)"
            raise ValueError(err_msg.format(lst=list_trim, v=var_zero_trim))

        if np.sum(keep_column_bool) == self._adata.shape[1]:
            pass
        else:
            Debug.vprint("Trimming {name} matrix {sh} to {n} columns".format(name=self.name, sh=self._adata.X.shape,
                                                                             n=np.sum(keep_column_bool)),
                         level=1)

            # This explicit copy allows the original to be deallocated
            # Otherwise the GC leaves the original because the view reference keeps it alive
            # At some point it will need to copy so why not now
            self._adata = AnnData(self._adata.X[:, keep_column_bool],
                                  obs=self._adata.obs.copy(),
                                  var=self._adata.var.loc[keep_column_bool, :].copy(),
                                  dtype=self._adata.X.dtype)

            # Make sure that there's no hanging reference to the original object
            gc.collect()

    def get_gene_data(self, gene_list, copy=False, force_dense=False, to_df=False, zscore=False):

        x = self._adata[:, gene_list]
        labels = x.var_names

        if (force_dense or to_df or zscore) and self.is_sparse:
            x = x.X.A
        else:
            x = x.X

        if zscore:
            x = np.subtract(x, self.obs_means.reshape(-1, 1))
            x = np.divide(x, self.obs_stdev.reshape(-1, 1))
        elif copy:
            x = x.copy()

        return pd.DataFrame(x, columns=labels, index=self.sample_names) if to_df else x

    def get_sample_data(self, sample_index, copy=False, force_dense=False, to_df=False, zscore=False):

        x = self._adata[sample_index, :]
        labels = x.obs_names

        if (force_dense or to_df or zscore) and self.is_sparse:
            x = x.X.A
        else:
            x = x.X

        if zscore:
            x = np.subtract(x, self.obs_means[sample_index].reshape(-1, 1))
            x = np.divide(x, self.obs_stdev[sample_index].reshape(-1, 1))
        elif copy:
            x = x.X.copy()

        return pd.DataFrame(x, columns=self.gene_names, index=labels) if to_df else x

    def get_bootstrap(self, sample_bootstrap_index):
        return InferelatorData(expression_data=self._adata[sample_bootstrap_index, :].X.copy(),
                               gene_names=self.gene_names)

    def subset_copy(self, row_index=None, column_index=None):

        if row_index is not None and column_index is not None:
            data_view = self._adata[row_index, column_index]
        elif row_index is not None:
            data_view = self._adata[row_index, :]
        elif column_index is not None:
            data_view = self._adata[: column_index]
        else:
            data_view = self._adata

        return InferelatorData(data_view.copy())

    def dot(self, other, other_is_right_side=True, force_dense=False):
        """
        Calculate dot product
        :param other:
        :param other_is_right_side:
        :param force_dense:
        :return:
        """

        if other_is_right_side:
            return DotProduct.dot(self._adata.X, other, cast=True, dense=force_dense)
        else:
            return DotProduct.dot(other, self._adata.X, cast=True, dense=force_dense)

    def to_csv(self, file_name, sep="\t"):

        if self.is_sparse:
            scipy.io.mmwrite(file_name, self.values)
        else:
            self._adata.to_df().to_csv(file_name, sep=sep)

    def to_h5ad(self, file_name, compression="gzip"):

        self._adata.write(file_name, compression=compression)

    def transform(self, func, add_pseudocount=False, memory_efficient=True, chunksize=1000):

        if add_pseudocount and self.is_sparse:
            self._adata.X.data += 1
        elif add_pseudocount:
            self._adata.X += 1

        if self.is_sparse:
            self._adata.X.data = func(self._adata.X.data)
        elif self._adata.X.ndim == 1 or self._is_integer:
            self._adata.X = func(self._adata.X)
        elif not memory_efficient and type(func(self._data.flat[0])) == self._adata.X.dtype:
            self._adata.X[...] = func(self._adata.X)
        elif memory_efficient and type(func(self._data.flat[0])) == self._adata.X.dtype:
            for i in range(math.ceil(self._adata.shape[0] / chunksize)):
                start, stop = i * chunksize, min(i + 1 * chunksize, self._adata.shape[0])
                self._adata.X[start:stop, :] = func(self._adata.X[start:stop, :])
        else:
            self._adata.X = func(self._adata.X)

    def add(self, val):
        """
        Add a value to the matrix in-place
        :param val: Value to add
        :type val: numeric
        """
        self._data[...] = self._data + val

    def subtract(self, val):
        """
        Subtract a value from the matrix in-place
        :param val: Value to subtract
        :type val: numeric
        """
        self._data[...] = self._data - val

    def divide(self, div_val, axis=None):
        """
        Divide a value from the matrix in-place
        :param div_val: Value to divide
        :type div_val: numeric
        :param axis: Which axis to divide from (0, 1, or None)
        :type axis: int, None
        """

        if self._is_integer:
            self.convert_to_float()

        if self.is_sparse and axis is None:
            self._adata.X.data /= div_val
        elif self.is_sparse and ((sparse.isspmatrix_csr(self._adata.X) and axis == 1) or
                                 (sparse.isspmatrix_csc(self._adata.X) and axis == 0)):
            if not hasattr(div_val, "ndim") or div_val.ndim != 1 or self.shape[0 if axis else 1] != div_val.shape[0]:
                raise ValueError("Division array is not aligned")
            self._adata.X.data /= np.repeat(div_val, self._adata.X.getnnz(axis=axis))
        elif self.is_sparse:
            raise ValueError("axis = 1 only works for CSC & axis = 0 only works for CSR")
        elif axis is None:
            self._adata.X /= div_val
        elif axis == 0:
            self._adata.X /= div_val[None, :]
        elif axis == 1:
            self._adata.X /= div_val[:, None]
        else:
            raise ValueError("axis must be 0, 1 or None")

    def multiply(self, mult_val, axis=None):
        """
        Multiply the matrix by a value in-place
        :param mult_val: Value to multiply
        :type mult_val: numeric
        :param axis: Which axis to divide from (0, 1, or None)
        :type axis: int, None
        """

        if self._is_integer:
            self.convert_to_float()

        if self.is_sparse and axis is None:
            self._adata.X.data *= mult_val
        elif self.is_sparse and ((sparse.isspmatrix_csr(self._adata.X) and axis == 1) or
                                 (sparse.isspmatrix_csc(self._adata.X) and axis == 0)):
            if not hasattr(mult_val, "ndim") or mult_val.ndim != 1 or self.shape[0 if axis else 1] != mult_val.shape[0]:
                raise ValueError("Division array is not aligned")
            self._adata.X.data *= np.repeat(mult_val, self._adata.X.getnnz(axis=axis))
        elif self.is_sparse:
            raise ValueError("axis = 1 only works for CSC & axis = 0 only works for CSR")
        elif axis is None:
            self._adata.X *= mult_val
        elif axis == 0:
            self._adata.X *= mult_val[None, :]
        elif axis == 1:
            self._adata.X *= mult_val[:, None]
        else:
            raise ValueError("axis must be 0, 1 or None")

    def zscore(self, axis=0, ddof=1):

        self.convert_to_float()
        self.to_dense()

        if axis == 0:
            for i in range(self.shape[1]):
                self._data[:, i] = scale_vector(self._data[:, i], ddof=ddof)
        elif axis == 1:
            for i in range(self.shape[0]):
                self._data[i, :] = scale_vector(self._data[i, :], ddof=ddof)

        return self

    def copy(self):

        new_data = InferelatorData(self.values.copy(),
                                   meta_data=self.meta_data.copy(),
                                   gene_data=self.gene_data.copy())

        new_data._adata.var_names = cp.copy(self._adata.var_names)
        new_data._adata.obs_names = cp.copy(self._adata.obs_names)
        new_data._adata.uns = cp.copy(self._adata.uns)

        return new_data

    def to_csc(self):

        if self.is_sparse and not sparse.isspmatrix_csc(self._adata.X):
            self._adata.X = sparse.csc_matrix(self._adata.X)

    def to_csr(self):

        if self.is_sparse and not sparse.isspmatrix_csr(self._adata.X):
            self._adata.X = sparse.csr_matrix(self._adata.X)

    def to_dense(self):

        if self.is_sparse:
            self._adata.X = self._adata.X.A

    def to_sparse(self, mode="csr"):

        if not self.is_sparse and mode.lower() == "csr":
            self._adata.X = sparse.csr_matrix(self._adata.X)
        elif not self.is_sparse and mode.lower() == "csc":
            self._adata.X = sparse.csc_matrix(self._adata.X)
        elif not self.is_sparse:
            raise ValueError("Mode must be csc or csr")

    def to_df(self):
        return self._adata.to_df()

    @staticmethod
    def _make_idx_str(df):
        df.index = df.index.astype(str) if not pat.is_string_dtype(df.index.dtype) else df.index
        df.columns = df.columns.astype(str) if not pat.is_string_dtype(df.columns.dtype) else df.columns



import pandas as pd
import scipy.sparse as sps
from inferelator.utils import InferelatorData


class TestDataSingleCellLike(object):
    expression_matrix = pd.DataFrame([[2, 28, 0, 16, 1, 3], [6, 21, 0, 3, 1, 3], [4, 39, 0, 17, 1, 3],
                                      [8, 34, 0, 7, 1, 3], [6, 26, 0, 3, 1, 3], [1, 31, 0, 1, 1, 4],
                                      [3, 27, 0, 5, 1, 4], [8, 34, 0, 9, 1, 3], [1, 22, 0, 3, 1, 4],
                                      [9, 33, 0, 17, 1, 2]],
                                     columns=["gene1", "gene2", "gene3", "gene4", "gene5", "gene6"]).transpose()
    meta_data = pd.DataFrame({"Condition": ["A", "B", "C", "C", "B", "B", "A", "C", "B", "C"],
                              "Genotype": ['WT', 'WT', 'WT', 'WT', 'WT', 'WT', 'WT', 'WT', 'WT', 'WT']})
    priors_data = pd.DataFrame([[0, 1], [0, 1], [1, 0], [0, 0]], index=["gene1", "gene2", "gene4", "gene5"],
                               columns=["gene3", "gene6"])
    gene_metadata = pd.DataFrame({"SystematicName": ["gene1", "gene2", "gene3", "gene4", "gene7", "gene6"]})
    gene_list_index = "SystematicName"
    tf_names = ["gene3", "gene6"]


TEST_DATA = InferelatorData(TestDataSingleCellLike.expression_matrix,
                            transpose_expression=True,
                            meta_data=TestDataSingleCellLike.meta_data,
                            gene_data=TestDataSingleCellLike.gene_metadata,
                            gene_data_idx_column=TestDataSingleCellLike.gene_list_index)

TEST_DATA_SPARSE = InferelatorData(sps.csr_matrix(TestDataSingleCellLike.expression_matrix.T.values),
                                   gene_names=TestDataSingleCellLike.expression_matrix.index,
                                   sample_names=TestDataSingleCellLike.expression_matrix.columns,
                                   meta_data=TestDataSingleCellLike.meta_data,
                                   gene_data=TestDataSingleCellLike.gene_metadata,
                                   gene_data_idx_column=TestDataSingleCellLike.gene_list_index)

CORRECT_GENES_INTERSECT = pd.Index(["gene1", "gene2", "gene3", "gene4", "gene6"])
CORRECT_GENES_NZ_VAR = pd.Index(["gene1", "gene2", "gene4", "gene6"])



import numpy as np
import pandas as pd
import scipy.stats
import copy

from inferelator.utils import Debug, InferelatorData
from inferelator.distributed.inferelator_mp import MPControl
from inferelator.utils import Validator as check

DEFAULT_CHUNK = 25
PROGRESS_STR = "Regression on {gn} [{i} / {total}]"


class BaseRegression(object):
    # These are all the things that have to be set in a new regression class

    chunk = DEFAULT_CHUNK  # int

    # Raw Data
    X = None  # [K x N] float
    Y = None  # [G x N] float
    G = None  # int G
    K = None  # int K

    def __init__(self, X, Y):
        """
        Create a regression object and do basic data transforms

        :param X: Expression or Activity data [N x K]
        :type X: InferelatorData
        :param Y: Response expression data [N x G]
        :type Y: InferelatorData
        """

        # Get the IDs and total count for the genes and predictors
        self.K = X.num_genes
        self.tfs = X.gene_names
        self.G = Y.num_genes
        self.genes = Y.gene_names

        # Rescale the design expression or activity data on features
        self.X = X
        self.X.zscore()

        self.Y = Y

        Debug.vprint("Predictor matrix {pr} and response matrix {re} ready".format(pr=X.shape, re=Y.shape))

    def run(self):
        """
        Execute regression separately on each response variable in the data

        :return: pd.DataFrame [G x K], pd.DataFrame [G x K]
            Returns the regression betas and beta error reductions for all threads if this is the master thread (rank 0)
            Returns None, None if it's a subordinate thread
        """

        run_data = self.regress()

        if MPControl.is_master:
            pileup_data = self.pileup_data(run_data)
        else:
            pileup_data = None, None

        MPControl.sync_processes("post_pileup")
        return pileup_data

    def regress(self):
        """
        Execute regression and return a list which can be provided to pileup_data
        :return: list
        """
        raise NotImplementedError

    def pileup_data(self, run_data):
        """
        Take the completed run data and pack it up into a DataFrame of betas

        :param run_data: list
            A list of regression result dicts ordered by gene. Each regression result should have `ind`, `pp`, `betas`
            and `betas_resc` keys with the appropriate data.
        :return betas, betas_rescale: (pd.DataFrame [G x K], pd.DataFrame [G x K])
        """

        # Create G x K arrays of 0s to populate with the regression data
        betas = np.zeros((self.G, self.K), dtype=np.dtype(float))
        betas_rescale = np.zeros((self.G, self.K), dtype=np.dtype(float))

        # Populate the zero arrays with the BBSR betas
        for data in run_data:

            # If data is None assume a null model
            if data is None:
                raise RuntimeError("No model produced by regression method")

            xidx = data['ind']  # Int
            yidx = data['pp']  # Boolean array of size K
            betas[xidx, yidx] = data['betas']
            betas_rescale[xidx, yidx] = data['betas_resc']

        d_len, b_avg, null_m = self._summary_stats(betas)
        Debug.vprint("Regression complete:", end=" ", level=0)
        Debug.vprint("{d_len} Models, {b_avg} Preds per Model ({nom} Null)".format(d_len=d_len,
                                                                                   b_avg=round(b_avg, 4),
                                                                                   nom=null_m), level=0)

        # Convert arrays into pd.DataFrames to return results
        betas = pd.DataFrame(betas, index=self.Y.gene_names, columns=self.X.gene_names)
        betas_rescale = pd.DataFrame(betas_rescale, index=self.Y.gene_names, columns=self.X.gene_names)

        return betas, betas_rescale

    @staticmethod
    def _summary_stats(arr):
        d_len = arr.shape[0]
        b_avg = np.mean(np.sum(arr != 0, axis=1))
        null_m = np.sum(np.sum(arr != 0, axis=1) == 0)
        return d_len, b_avg, null_m


class _RegressionWorkflowMixin(object):
    """
    RegressionWorkflow implements run_regression and run_bootstrap
    Each regression method needs to extend this to implement run_bootstrap (and also run_regression if necessary)
    """

    def set_regression_parameters(self, **kwargs):
        """
        Set any parameters which are specific to one or another regression method
        """
        pass

    def run_regression(self):
        betas = []
        rescaled_betas = []

        MPControl.sync_processes("pre_regression")

        for idx, bootstrap in enumerate(self.get_bootstraps()):
            Debug.vprint('Bootstrap {} of {}'.format((idx + 1), self.num_bootstraps), level=0)
            np.random.seed(self.random_seed + idx)
            current_betas, current_rescaled_betas = self.run_bootstrap(bootstrap)
            if self.is_master():
                betas.append(current_betas)
                rescaled_betas.append(current_rescaled_betas)

            MPControl.sync_processes("post_bootstrap")

        return betas, rescaled_betas

    def run_bootstrap(self, bootstrap):
        raise NotImplementedError


class _MultitaskRegressionWorkflowMixin(_RegressionWorkflowMixin):
    """
    MultitaskRegressionWorkflow implements run_regression and run_bootstrap for multitask workflow
    Each regression method needs to extend this to implement run_bootstrap (and also run_regression if necessary)
    """

    def run_regression(self):

        betas = [[] for _ in range(self._n_tasks)]
        rescaled_betas = [[] for _ in range(self._n_tasks)]

        for idx in range(self.num_bootstraps):
            Debug.vprint('Bootstrap {} of {}'.format((idx + 1), self.num_bootstraps), level=0)
            current_betas, current_rescaled_betas = self.run_bootstrap(idx)

            if self.is_master():
                for k in range(self._n_tasks):
                    betas[k].append(current_betas[k])
                    rescaled_betas[k].append(current_rescaled_betas[k])

        return betas, rescaled_betas

    def run_bootstrap(self, bootstrap):
        raise NotImplementedError


def recalculate_betas_from_selected(x, y, idx=None):
    """
    Estimate betas from a selected subset of predictors
    :param x: np.ndarray [N x k]
        Predictor matrix
    :param y: np.ndarray [N x 1]
        Response matrix
    :param idx: np.ndarray [k x 1]
        Predictors to use (unused predictors will return a beta of 0)
        If None, use all predictors
    :return: np.ndarray [k,]
        Estimated beta-hats
    """

    # Create an array of size [k,] to hold the estimated betas
    best_betas = np.zeros(x.shape[1], dtype=np.dtype(float))

    # Use all predictors if no subset index is passed in
    if idx is None:
        idx = np.ones(x.shape[1], dtype=np.dtype(bool))

    # Convert boolean array to an array of indexes
    idx = bool_to_index(idx)

    # Subset the predictors with the index array
    x = x[:, idx]

    # Solve for beta-hat with LAPACK or return a null model if xTx is singular
    xtx = np.dot(x.T, x)
    if np.linalg.matrix_rank(xtx) == xtx.shape[1]:
        beta_hat = np.linalg.solve(np.dot(x.T, x), np.dot(x.T, y))
    else:
        beta_hat = np.zeros(len(idx), dtype=np.dtype(float))

    # Use the index array to write beta-hats
    # This yields the same size result matrix as number of predictors in x
    # (even if x is subset with an index)
    for i, j in enumerate(idx):
        best_betas[j] = beta_hat[i]
    return best_betas


def predict_error_reduction(x, y, betas):
    """
    Predict the error reduction from each predictor
    :param x: np.ndarray [n x k]
    :param y: np.ndarray [n x 1]
    :param betas: np.ndarray [k x 1]
    :return: np.ndarray [k,]
    """
    assert check.argument_type(betas, np.ndarray)

    (n, k) = x.shape
    pp_idx = index_of_nonzeros(betas).tolist()

    # Calculate the variance of the residuals
    ss_all = sigma_squared(x, y, betas)
    error_reduction = np.zeros(k, dtype=np.dtype(float))

    if len(pp_idx) == 1:
        error_reduction[pp_idx] = 1 - (ss_all / np.var(y, ddof=1))
        return error_reduction

    for pp_i in range(len(pp_idx)):
        # Copy the index of predictors
        leave_out = copy.copy(pp_idx)
        # Pull off one of the predictors
        lost = leave_out.pop(pp_i)

        # Reestimate betas for all the predictors except the one that we removed
        x_leaveout = x[:, leave_out]
        try:
            xt = x_leaveout.T
            xtx = np.dot(xt, x_leaveout)
            xty = np.dot(xt, y)
            beta_hat = scipy.linalg.solve(xtx, xty, assume_a='sym')
        except np.linalg.LinAlgError:
            beta_hat = np.zeros(len(leave_out), dtype=np.dtype(float))

        # Calculate the variance of the residuals for the new estimated betas
        ss_leaveout = sigma_squared(x_leaveout, y, beta_hat)

        # Check to make sure that the ss_all and ss_leaveout differences aren't just precision-related
        if np.abs(ss_all - ss_leaveout) < np.finfo(float).eps * len(pp_idx):
            error_reduction[lost] = 0.
        else:
            error_reduction[lost] = 1 - (ss_all / ss_leaveout)

    return error_reduction


def sigma_squared(x, y, betas):
    return np.var(np.subtract(y, np.dot(x, betas).reshape(-1, 1)), ddof=1)


def index_of_nonzeros(arr):
    """
    Returns an array that indexes all the non-zero elements of an array
    :param arr: np.ndarray
    :return: np.ndarray
    """
    return np.where(arr != 0)[0]


def bool_to_index(arr):
    """
    Returns an array that indexes all the True elements of a boolean array
    :param arr: np.ndarray
    :return: np.ndarray
    """
    assert check.argument_type(arr, np.ndarray)
    return np.where(arr)[0]



import numpy as np
import pandas as pd
import os

from inferelator import utils
from inferelator.utils import Validator as check
from inferelator.postprocessing.model_performance import RankSummingMetric
from inferelator.postprocessing import (TARGET_COLUMN, REGULATOR_COLUMN, PRECISION_COLUMN, RECALL_COLUMN,
                                        CONFIDENCE_COLUMN, GOLD_STANDARD_COLUMN)

import matplotlib
matplotlib.use('pdf')
import matplotlib.pyplot as plt


class RankSummaryPR(RankSummingMetric):
    """
    This class extends RankSumming and calculates precision-recall
    """

    name = "AUPR"
    curve_file_name = "pr_curve.pdf"

    # PR
    @property
    def aupr(self):
        return self.calculate_aupr(self.filtered_data)

    def __init__(self, rankable_data, gold_standard, filter_method='keep_all_gold_standard'):

        super(RankSummaryPR, self).__init__(rankable_data, gold_standard, filter_method=filter_method)

        # Calculate the precision and recall and store them with confidence data
        self.filtered_data = self.calculate_precision_recall(self.filtered_data.copy())

        # Join the filtered precision/recall onto the full confidences
        join_data = self.filtered_data.loc[:, [TARGET_COLUMN, REGULATOR_COLUMN, PRECISION_COLUMN, RECALL_COLUMN]]
        join_data = join_data.set_index([TARGET_COLUMN, REGULATOR_COLUMN])
        self.confidence_data = self.confidence_data.join(join_data, on=[TARGET_COLUMN, REGULATOR_COLUMN])

    def score(self):

        return self.name, self.aupr

    def auc(self):

        return self.aupr

    def curve_dataframe(self):

        return self.filtered_data.loc[:, [PRECISION_COLUMN, RECALL_COLUMN]]

    def output_curve(self, ax=None, figsize=(6, 4)):
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)
            ax = fig.add_subplot(1, 1, 1)

        # Extract the recall and precision data
        recall, precision = self.modify_pr(self.curve_dataframe())

        self.plot_pr_curve(recall, precision, self.aupr, ax)

        return ax

    @staticmethod
    def plot_pr_curve(recall, precision, aupr, ax):

        # Generate a plot
        ax.plot(recall, precision)
        ax.set_xlabel('Recall')
        ax.set_ylabel('Precision')
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

        # Add the AUPR as an annotation
        ax.annotate("aupr = {aupr:.5f}".format(aupr=aupr), xy=(0.4, 0.05), xycoords='axes fraction')

        return ax

    def num_over_precision_threshold(self, threshold):
        return np.sum(self.confidence_data[CONFIDENCE_COLUMN] >= self.find_threshold(PRECISION_COLUMN, threshold))

    def num_over_recall_threshold(self, threshold):
        return np.sum(self.confidence_data[CONFIDENCE_COLUMN] >= self.find_threshold(PRECISION_COLUMN, threshold))

    def num_over_conf_threshold(self, threshold):
        return np.sum(self.confidence_data[CONFIDENCE_COLUMN] >= threshold)

    def find_threshold(self, column_name, threshold):

        assert check.argument_numeric(threshold, low=0, high=1)

        threshold_index = self.confidence_data[column_name] >= threshold

        # If there's nothing in the index return np.inf.
        if np.sum(threshold_index) == 0:
            return np.inf
        else:
            return self.confidence_data.loc[threshold_index, CONFIDENCE_COLUMN].min()

    @staticmethod
    def calculate_precision_recall(data, transform_ties=None):
        """
        Calculate the precision & recall based on the confidence scores and gold standard
        :param data: pd.DataFrame
            Dataframe with a gold standard and confidence column, sorted on confidence column
        :return data: pd.DataFrame
            Sorted dataframe with additional precision and recall columns
        """

        # Make sure data is sorted
        if not data.loc[~pd.isnull(data[CONFIDENCE_COLUMN]), CONFIDENCE_COLUMN].is_monotonic_decreasing:
            data = data.sort_values(by=CONFIDENCE_COLUMN, ascending=False, na_position='last')
            data = data.reset_index()
            utils.Debug.vprint("Resorting confidences for PR", level=0)

        # Get indices for stuff
        valid_gs_idx = ~pd.isnull(data[GOLD_STANDARD_COLUMN])

        # Find the edges that are in the gold standard
        valid_gs = (data.loc[valid_gs_idx, GOLD_STANDARD_COLUMN] != 0).astype(int)

        # the following mimics the R function ChristophsPR
        # Add nan columns
        data[PRECISION_COLUMN] = np.nan
        data[RECALL_COLUMN] = np.nan

        # Calculate precision [TP / (TP + FP)]
        data.loc[valid_gs_idx, PRECISION_COLUMN] = np.cumsum(valid_gs).astype(float) / np.arange(1, len(valid_gs)+1, 1)

        # Calculate recall [TP / (TP + FN)]
        data.loc[valid_gs_idx, RECALL_COLUMN] = np.cumsum(valid_gs).astype(float) / sum(valid_gs)

        if transform_ties is not None:
            RankSummingMetric.transform_column(data, CONFIDENCE_COLUMN, PRECISION_COLUMN, transform_ties)
            RankSummingMetric.transform_column(data, CONFIDENCE_COLUMN, RECALL_COLUMN, transform_ties)

        else:
            # Overwrite the precision of no-confidence with the mean value
            zero_confidence = data[CONFIDENCE_COLUMN] == 0
            zero_confidence_precision_idx = zero_confidence & valid_gs_idx

            zero_confidence_precision_val = data.loc[zero_confidence_precision_idx, PRECISION_COLUMN].mean()
            data.loc[zero_confidence_precision_idx, PRECISION_COLUMN] = zero_confidence_precision_val

        return data

    @staticmethod
    def modify_pr(data):
        """
        Inserts values into the precision and recall to allow for plotting & calculations of area
        :param data: pd.DataFrame
            Sorted confidences with a PRECISION and a RECALL column
        :return precision: np.ndarray
            Precision values
        :return recall: np.ndarray
            Recall values
        """

        data = data.loc[~pd.isnull(data[PRECISION_COLUMN]), :]
        precision = np.insert(data[PRECISION_COLUMN].values, 0, data[PRECISION_COLUMN].iloc[0])
        recall = np.insert(data[RECALL_COLUMN].values, 0, 0)
        return recall, precision

    @staticmethod
    def calculate_aupr(data):
        recall, precision = RankSummaryPR.modify_pr(data)
        # using midpoint integration to calculate the area under the curve
        d_recall = np.diff(recall)
        m_precision = precision[:-1] + np.diff(precision) / 2
        return sum(d_recall * m_precision)


"""
MultiprocessingController runs everything through a multiprocessing Pool
This requires pathos because the default multiprocessing serializes with cPickle
"""

import pathos
import collections

from inferelator.distributed import AbstractController
from inferelator.utils import Validator as check


class MultiprocessingController(AbstractController):
    _controller_name = "multiprocessing"
    client = None
    is_master = True

    # Control variables
    chunk = 25

    # Num processes
    processes = 4

    @classmethod
    def connect(cls, *args, **kwargs):
        cls.client = pathos.multiprocessing.ProcessPool(nodes=cls.processes, **kwargs)
        return True

    @classmethod
    def sync_processes(cls, *args, **kwargs):
        return True

    @classmethod
    def set_processes(cls, process_count):
        """
        Set the number of dask workers to use
        :param process_count: int
        :return:
        """
        check.argument_integer(process_count, low=1)

        cls.processes = process_count

    @classmethod
    def map(cls, func, *args, **kwargs):
        """
        Map a function across iterable(s) and return a list of results

        :param func: function
            Mappable function
        :param args: iterable
            Iterator(s)
        """
        assert check.argument_callable(func)
        assert check.argument_list_type(args, collections.Iterable)
        return cls.client.map(func, *args, chunksize=cls.chunk)

    @classmethod
    def shutdown(cls):
        return cls.client.close()
